///////////////////////////////////////////////////////////////////////////////
//                             LCD_MM.c                                      //
//                 For LCD modules based on HD44780                          //
//                                                                           //
// lcd_init()        Initialize the LCD.                                     //
//                                                                           //
// lcd_putc(c)       Displays the character "c"; may use the ASCII code.     //
//                   Special characters:                                     //
//                                        \f  Clear display                  //
//                                        \n  Go to start of second line     //
//                                        \b  Move back one position         //
//                                                                           //
// lcd_gotoxy(x,y)   Set cursor position (column, line).                     //
//                                                                           //
// lcd_getc(x,y)     Returns the character at position x,y.                  //
//                                                                           //
// lcd_blink(b)      0 or 1; blinks the character at the cursor position.    //
//                                                                           //
// lcd_cursor(c)     0 or 1; makes the cursor visible at the cursor position.//
//                                                                           //
// Note: be aware that the use of lcd_blink(b) and lcd_cursor(c) cannot be   //
// accomplished simultaneously, since one cancels the other.                 //
///////////////////////////////////////////////////////////////////////////////

// The pin connections must be done according to the following table:
//     PIC        LCD
//    PORTX1 ---- rs
//    PORTX2 ---- rw
//    PORTX3 ---- en
//    PORTX4 ---- D4
//    PORTX5 ---- D5
//    PORTX6 ---- D6
//    PORTX7 ---- D7
//
// PORTX refers to the selected PIC port; the number is the corresponding bit.
// The PORTX0 at the PIC side is not used; therefore, the external INT (PORTB0)
// will be available if the port B is selected.
// LCD pins D0-D3 are not used, since the 4-bit mode will be selected.

struct lcd_pin_map {          
           int1 unused;       
           int1 rs;           
           int1 rw;           
           int1 en;           
           int  data:4;         
        } lcd;

// Select the port to use by changing the last letter of PORTB and set_tris_b.
#locate lcd=getenv("sfr:PORTB")    
#define set_tris_lcd(x) set_tris_b(x)

#define lcd_type 2            // 0=5x7, 1=5x10, 2=2 lines
#define lcd_line_two 0x40    

BYTE const LCD_INIT_STRING[4]={0x20 | (lcd_type << 2), 0xc, 1, 6};
                          
struct lcd_pin_map const LCD_WRITE={1,0,0,0,0};
struct lcd_pin_map const LCD_READ={1,0,0,0,15};

BYTE lcd_read_byte() {
      BYTE low,high;
      set_tris_lcd(LCD_READ);
      lcd.rw=1;
      delay_cycles(1);
      lcd.en=1;
      delay_cycles(1);
      high=lcd.data;
      lcd.en=0;
      delay_cycles(1);
      lcd.en=1;
      delay_us(1);
      low=lcd.data;
      lcd.en=0;
      set_tris_lcd(LCD_WRITE);
      return((high<<4) | low);
}

void lcd_send_nibble(BYTE n) {
      lcd.data=n;
      delay_cycles(1);
      lcd.en=1;
      delay_us(2);
      lcd.en=0;
}

void lcd_send_byte(BYTE address,BYTE n) {
      lcd.rs=0;
      while(bit_test(lcd_read_byte(),7)) ;
      lcd.rs=address;
      delay_cycles(1);
      lcd.rw=0;
      delay_cycles(1);
      lcd.en=0;
      lcd_send_nibble(n >> 4);
      lcd_send_nibble(n & 0xf);
}

void lcd_init() {
    BYTE i;
    set_tris_lcd(LCD_WRITE);
    lcd.rs = 0;
    lcd.rw = 0;
    lcd.en = 0;
    delay_ms(15);
    for(i=1;i<=3;++i) {
       lcd_send_nibble(3);
       delay_ms(5);
    }
    lcd_send_nibble(2);
    for(i=0;i<=3;++i)
       lcd_send_byte(0,LCD_INIT_STRING[i]);
}

void lcd_gotoxy(BYTE x,BYTE y) {
   BYTE address;
   if(y!=1)
      address=lcd_line_two;
   else
      address=0;
   address+=x-1;
   lcd_send_byte(0,0x80|address);
}

void lcd_putc(char c) {
   switch (c) {
      case '\f'   : lcd_send_byte(0,1);
                    delay_ms(2);            break;
      case '\n'   : lcd_gotoxy(1,2);        break;
      case '\b'   : lcd_send_byte(0,0x10);  break;
      default     : lcd_send_byte(1,c);     break;
   }
}

char lcd_getc(BYTE x,BYTE y) {
    char value;
    lcd_gotoxy(x,y);
    while(bit_test(lcd_read_byte(),7));
    lcd.rs=1;
    value=lcd_read_byte();
    lcd.rs=0;
    return(value);
}

void lcd_blink(int1 b) {
   switch (b) {
      case 0   : lcd_send_byte(0,0b00001100);
                                          break;
      case 1   : lcd_send_byte(0,0b00001101);
                                          break;
   }
}

void lcd_cursor(int1 c) {
   switch (c) {
      case 0   : lcd_send_byte(0,0b00001100);
                                          break;
      case 1   : lcd_send_byte(0,0b00001110);
                                          break;
   }
}
